from contextlib import asynccontextmanager, contextmanager
from contextvars import ContextVar
from typing import Any

from fastapi import Request

_GLOBAL_REQUEST_STORE: ContextVar[dict[Any, Any]] = ContextVar(
    '_GLOBAL_REQUEST_STORE',
    default={},
)


class RequestScopeService:
    CONTEXT_MANAGERS_KEY = '__Context_Managers__'
    CURRENT_REQUEST_KEY = '__Current_Request__'

    @classmethod
    def get_request_store(cls):
        return _GLOBAL_REQUEST_STORE.get()

    @classmethod
    def get_request_key(cls, item):
        return item

    @classmethod
    def set_current_request(cls, request: Request):
        store = cls.get_request_store()
        store[cls.CURRENT_REQUEST_KEY] = request

    @classmethod
    def get_current_request(cls) -> Request:
        store = cls.get_request_store()

        request = store.get(cls.CURRENT_REQUEST_KEY)

        if request is None:
            raise RuntimeError(
                'No Request instance found in the current request scope. '
                'This indicates that the RequestScopeMiddleware did not run before '
                'dependency resolution.\n\n'
                'Most common causes:\n'
                '  • The middleware was not added: app.add_middleware(RequestScopeMiddleware)\n'
                '  • The container is resolving dependencies outside an HTTP request\n'
                '  • A background task is trying to access Request\n\n'
                'How to fix:\n'
                '  Ensure that RequestScopeMiddleware is registered before any routes and that '
                'all Request-dependent dependencies are only resolved inside an HTTP request.'
            )

        return request

    @classmethod
    @contextmanager
    def request_scope(cls):
        token = cls._start_request_scope()

        try:
            yield cls
        except Exception:
            import sys

            cls._exit_request_scope(token, sys.exc_info())
            raise
        else:
            cls._exit_request_scope(token)

    @classmethod
    @asynccontextmanager
    async def request_scope_async(cls):
        token = cls._start_request_scope()

        try:
            yield cls
        except Exception:
            import sys

            await cls._exit_request_scope_async(token, sys.exc_info())
            raise
        else:
            await cls._exit_request_scope_async(token)

    @classmethod
    def _start_request_scope(cls):
        token = _GLOBAL_REQUEST_STORE.set({})
        return token

    @classmethod
    def _exit_request_scope(cls, token, exc_info=None):
        store = _GLOBAL_REQUEST_STORE.get(token)

        context_managers = store.get(cls.CONTEXT_MANAGERS_KEY, [])

        if exc_info and exc_info[0] is not None:
            for cm in reversed(context_managers):
                try:
                    if hasattr(cm, '__exit__'):
                        cm.__exit__(*exc_info)
                    elif hasattr(cm, 'close'):
                        cm.close()

                except Exception:
                    pass
        else:
            for cm in reversed(context_managers):
                try:
                    if hasattr(cm, '__exit__'):
                        cm.__exit__(None, None, None)
                    elif hasattr(cm, 'close'):
                        cm.close()

                except Exception:
                    pass

        _GLOBAL_REQUEST_STORE.reset(token)

    @classmethod
    async def _exit_request_scope_async(cls, token, exc_info=None):
        store = _GLOBAL_REQUEST_STORE.get()
        context_managers = store.get(cls.CONTEXT_MANAGERS_KEY, [])

        if exc_info and exc_info[0] is not None:
            for cm in reversed(context_managers):
                try:
                    if hasattr(cm, '__aexit__'):
                        await cm.__aexit__(*exc_info)
                    elif hasattr(cm, '__exit__'):
                        cm.__exit__(*exc_info)
                    elif hasattr(cm, 'aclose'):
                        await cm.aclose()
                    elif hasattr(cm, 'close'):
                        cm.close()
                except Exception:
                    pass
        else:
            for cm in reversed(context_managers):
                try:
                    if hasattr(cm, '__aexit__'):
                        await cm.__aexit__(None, None, None)
                    elif hasattr(cm, '__exit__'):
                        cm.__exit__(None, None, None)
                    elif hasattr(cm, 'aclose'):
                        await cm.aclose()
                    elif hasattr(cm, 'close'):
                        cm.close()
                except Exception:
                    pass

        _GLOBAL_REQUEST_STORE.reset(token)



import contextlib


def wrap_sync_gen(gen_fn, params):
    @contextlib.contextmanager
    def _ctx():
        gen = gen_fn(**params)
        exception_to_raise = None
        try:
            value = next(gen)
            yield value
        except Exception as e:
            exception_to_raise = e
        finally:
            try:
                if exception_to_raise:
                    gen.throw(type(exception_to_raise), exception_to_raise, exception_to_raise.__traceback__)
                else:
                    next(gen)
            except StopIteration:
                pass
            except Exception:
                if exception_to_raise:
                    raise exception_to_raise
                raise

    return _ctx()


@contextlib.asynccontextmanager
async def wrap_async_gen(gen_fn, params):
    gen = gen_fn(**params)
    exception_to_raise = None
    try:
        value = await gen.__anext__()
        yield value
    except Exception as e:
        exception_to_raise = e
    finally:
        try:
            if exception_to_raise:
                await gen.athrow(type(exception_to_raise), exception_to_raise, exception_to_raise.__traceback__)
            else:
                await gen.__anext__()
        except StopAsyncIteration:
            pass
        except Exception:
            if exception_to_raise:
                raise exception_to_raise
            raise



from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Any, AsyncGenerator, Awaitable, Generator

type Resolvable[T] = (
    Callable[..., T]
    | Callable[..., Awaitable[T]]
    | Callable[..., AsyncGenerator[T, Any]]
    | Callable[..., Generator[T, Any, Any]]
)

type Provider[T] = Callable[[], T] | Callable[[], Awaitable[T]]
type ProviderSource[T = Any] = type[T] | Resolvable[T]


class Request[T]: ...


class Singleton[T]: ...


class Transient[T]: ...


class Scope(Enum):
    SINGLETON = 'singleton'
    TRANSIENT = 'transient'
    REQUEST = 'request'


@dataclass
class Token:
    name: str


@dataclass
class ProviderInfo[T = Any]:
    provider: Provider[T]
    source: ProviderSource[T]
    needs_async: bool
    scope: Scope


class ProviderDependency:
    def __init__(self, provider_source: ProviderSource[Any]) -> None:
        self.provider_source = provider_source



from io import StringIO
from typing import Any, Literal


class MissingProviderError(RuntimeError):
    pass


class UnexpectedCoroutineError(RuntimeError):
    pass


class CircularDependencyError(Exception):
    def __init__(self, visited: dict[Any, Literal[True]], source: Any) -> None:
        dependency_graph_string = self._get_dependency_graph_string(visited, source)
        super().__init__(f'Circular dependency detected:\n{dependency_graph_string}')

    def _get_dependency_graph_string(self, visited: dict[Any, Literal[True]], source: Any) -> str:
        builder = StringIO()

        visited_sources = list(visited.keys()) + [str(source)]

        for i, v in enumerate(visited_sources):
            tab_level = i * 4

            if i == 0:
                builder.write('\n')

            if i != 0:
                builder.write(' ' * tab_level + ':\n')
                builder.write(' ' * tab_level + ':\n')

            if i == len(visited):
                builder.write(f'{" " * tab_level} {v} \x1b[31m  <- Circular dependency detected!\x1b[0m\n\n')
            else:
                builder.write(f'{" " * tab_level} {v}\n')

        return builder.getvalue()
